#!/bin/sh

# Copyright (C) 2006 Kel Modderman <kelrin@tpg.com.au>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# On Debian GNU/Linux systems, the text of the GPL license can be
# found in /usr/share/common-licenses/GPL.

if [ -n "$IF_WPA_MAINT_DEBUG" ]; then
	set -x
fi

# quit if we're called for the loopback
if [ "$IFACE" = lo ]; then
	exit 0
fi

WPA_SUP_BIN="/sbin/wpa_supplicant"
WPA_SUP_PNAME="wpa_supplicant"
WPA_SUP_PIDFILE="/var/run/wpa_supplicant.$IFACE.pid"

WPA_CLI_BIN="/sbin/wpa_cli"
WPA_CLI_PNAME="wpa_cli"
WPA_CLI_PIDFILE="/var/run/wpa_action.$IFACE.pid"
WPA_CLI_LOCKFILE="/var/lock/wpa_action.$IFACE.lock"

# path to common ctrl_interface socket for active interfaces
# this is the default path provided by upstream
WPA_CTRL_DIR="/var/run/wpa_supplicant"

# quit if executables are not installed
if [ ! -x "$WPA_SUP_BIN" -o ! -x "$WPA_CLI_BIN" ]; then
	exit 0
fi

if [ -n "$IF_WPA_VERBOSITY" -o "$VERBOSITY" = "1" ]; then
	# show wpa_cli messages when VERBOSITY has been set
	TO_NULL="/dev/stdout"
	# start-stop-daemon verbosity
	DAEMON_VERBOSITY="--verbose"
else
	# hide infomational messages when VERBOSITY is not set
	TO_NULL="/dev/null"
	# start-stop-daemon is quiet
	DAEMON_VERBOSITY="--quiet"
fi

# error and verbose informational messages
wpa_msg () {
	case "$1" in 
		"verbose")
			# verbose message terminated with newline
			shift
			echo "$WPA_SUP_PNAME: $@" >$TO_NULL
			;;
		"action")
			# verbose action message with no trailing newline
			# useful for displaying wpa_cli feedback (OK|FAILED)
			shift
			echo -n "$WPA_SUP_PNAME: $@ -- " >$TO_NULL
			;;
		"stderr")
			# error message
			# commonly followed by an exit statement
			shift
			echo "$WPA_SUP_PNAME: $@...aborting!" >/dev/stderr
			;;
		*)
			;;
	esac
}

test_daemon_pidfile () {
	local DAEMON PIDFILE
	# daemon executable is first argument
	if [ -n "$1" ]; then
		DAEMON="$1"
	fi
	# daemon pidfile to be tested
	if [ -f "$2" ]; then
		PIDFILE="$2"
	fi
	# do the test if both exist
	if [ -n "$DAEMON" -a -f "$PIDFILE" ]; then
		# return 0 when $PIDFILE was created by $DAEMON
		# return 1 otherwise
		if start-stop-daemon --stop --quiet --signal 0 \
			--exec "$DAEMON" --pidfile "$PIDFILE"; then
			return 0
		else
			# pidfile is invalid, not created by $DAEMON
			wpa_msg verbose "removing stale pidfile \"$PIDFILE\""
			rm -f "$PIDFILE"
			return 1
		fi
	else
		return 1
	fi
}

# quit if wpa_action is active
if test_daemon_pidfile "$WPA_CLI_BIN" "$WPA_CLI_PIDFILE"; then
	exit 0
fi

if [ -n "$IF_WPA_ROAM" ]; then
	# wpa-roam is only funtional in conjunction with the 'manual' inet METHOD
	if [ "$METHOD" != "manual" ]; then
		wpa_msg stderr "wpa-roam can only be used with the \"manual\" inet METHOD"
		exit 1
	fi
	# mapping script must exist
	if [ -n "$IF_WPA_MAPPING_SCRIPT" ]; then
		if ! type "$IF_WPA_MAPPING_SCRIPT" >/dev/null; then
			wpa_msg stderr "mapping script \"$IF_WPA_MAPPING_SCRIPT\" is not valid"
			exit 1
		fi
	fi
	# if an external mapping script is to be used, one must be given
	if [ -n "$IF_WPA_MAPPING_SCRIPT_PRIORITY" -a -z "$IF_WPA_MAPPING_SCRIPT" ]; then
		wpa_msg stderr "cannot set \"wpa-mapping-script-priority 1\" without a valid \"wpa-mapping-script\""
		exit 1
	fi
	# shunt roam conf to IF_WPA_CONF
	IF_WPA_CONF="$IF_WPA_ROAM"
	# action script is wpa_action
	WPA_ACTION_SCRIPT="/sbin/wpa_action"
fi

# master function; determines if ifupdown.sh should do something or not
if [ -s "$IF_WPA_CONF" ]; then
	# get ctrl_interface socket path defined in conf
	WPA_SUP_CONF_CTRL_DIR=$(sed -n -e \
		's/[[:space:]]*#.*//g;s/[[:space:]]\+.*$//g;s/^\(ctrl_interface\|DIR\)=\(.*\)/\2/p' "$IF_WPA_CONF")
	# use ctrl_interface socket path when defined in user supplied conf, overriding the pre-set default
	if [ -n "$WPA_SUP_CONF_CTRL_DIR" ]; then
		WPA_CTRL_DIR="$WPA_SUP_CONF_CTRL_DIR"
	fi
	# always supply ctrl_interface socket path to daemon
	# it may not be created when not explicitly specified in IF_WPA_CONF but we depend on it
	WPA_SUP_CONF="-c $IF_WPA_CONF -C $WPA_CTRL_DIR"
elif set | grep -q "^IF_WPA"; then
	# any wpa-* line causes wpa_supplicant to launch with an unconfigured ctrl_interface socket
	# the socket is later configured by a series of set_network cmd's in conf_wpa_supplicant()
	WPA_SUP_CONF="-C $WPA_CTRL_DIR"
else
	exit 0
fi

# wpa_cli wrapper function
# avoid language related interferences
# supply path to common ctrl_interface socket(s)
# supply IFACE as argument
wpa_cli () {
	LC_ALL=C $WPA_CLI_BIN -p $WPA_CTRL_DIR -i $IFACE "$@"
}

# wpa_cli_do <value> <type> <variable> [set_network variable] <desc>
# process environment variables resulting from wpa-* lines in the interfaces file
# $1 = envorinment variable
# $2 = data type of variable {raw|ascii}
# $3 = wpa_cli variable, if $3 is set_network, shift and take set_network subvariable
# $4 = wpa-* string as it would appear in interfaces file, enhances information messages
wpa_cli_do () {
	
	if [ -z "$1" ]; then
		# environment variable not defined, no-op
		return
	fi
	
	local WPACLISET_VALUE WPACLISET_VARIABLE WPACLISET_DESC
	
	# provide a method to disable the forced choice of ascii data type
	# using this option requires that strings are quoted as required in interfaces file
	if [ -n "$IF_WPA_INPUT_TYPE" ]; then
		WPACLISET_VALUE="$1"
	else
		# wpa_cli value and type
		case "$2" in
			ascii)
				# Double quote
				WPACLISET_VALUE="\"$1\""
				;;
			raw|*)
				# Provide raw value
				WPACLISET_VALUE="$1"
				;;
		esac
	fi
	
	# wpa_cli variable
	case "$3" in
		set_network)
			if [ -z "$WPA_ID" ]; then
				# WPA_ID is required for set_network commands
				return
			fi
			# Shift to set_network variable
			shift
			WPACLISET_VARIABLE="set_network $WPA_ID $3"
			;;
		*)
			WPACLISET_VARIABLE="$3"
			;;
	esac
	
	# display description of wpa_cli command, censor sensitive knowledge
	case "$4" in
		*-psk|*-passphrase|*-passwd*|*-wep-key*)
			# For sensitive knowledge
			WPACLISET_DESC="$4 *****"
			;;
		*)
			# Echo actual value used
			WPACLISET_DESC="$4 $WPACLISET_VALUE"
			;;
	esac

	# echo debug/verbosity messages
	wpa_msg action "$WPACLISET_DESC"
	
	# NB: $WPACLISET_VARIABLE must be unprotected and $WPACLISET_VALUE must be protected
	wpa_cli $WPACLISET_VARIABLE "$WPACLISET_VALUE" >$TO_NULL
}

init_wpa_supplicant () {
	if [ ! -S "$WPA_CTRL_DIR/$IFACE" ]; then
		
		if [ -n "$IF_WPA_ACTION" ]; then
			wpa_msg stderr "wpa-action support has been removed, see wpa_action(8) manpage"
			return 1
		fi
		
		local WPA_SUP_DRIVER WPA_SUP_OPTIONS

		# action script sanity checking
		if [ -n "$WPA_ACTION_SCRIPT" ]; then
			if [ -x "$WPA_ACTION_SCRIPT" ]; then
				# wait for wpa_cli to attach
				WPA_SUP_OPTIONS="-W -B -P $WPA_SUP_PIDFILE -i $IFACE"
				wpa_msg verbose "wait for wpa_cli to attach"
			else
				wpa_msg stderr "action script \"$WPA_ACTION_SCRIPT\" not executable"
				return 1
			fi
		else
			# default options, daemonise, create pidfile, supply IFACE
			WPA_SUP_OPTIONS="-B -P $WPA_SUP_PIDFILE -i $IFACE"
		fi

		if [ -n "$IF_WPA_BRIDGE" ]; then
			WPA_SUP_OPTIONS="$WPA_SUP_OPTIONS -b $IF_WPA_BRIDGE"
			wpa_msg verbose "wpa-bridge $IF_WPA_BRIDGE"
		fi

		# driver type of interface, defaults to wext when undefined
		if [ -n "$IF_WPA_DRIVER" ]; then
			WPA_SUP_DRIVER="$IF_WPA_DRIVER"
			wpa_msg verbose "wpa-driver $WPA_SUP_DRIVER"
		else
			# wext is default/fallback
			WPA_SUP_DRIVER="wext"
			wpa_msg verbose "using default driver type: wpa-driver $WPA_SUP_DRIVER"
		fi

		# if we have passed the criteria, start wpa_supplicant
		if [ -n "$WPA_SUP_CONF" ]; then
			
			wpa_msg verbose "$WPA_SUP_BIN $WPA_SUP_OPTIONS -D $WPA_SUP_DRIVER $WPA_SUP_CONF"
			
			start-stop-daemon --start --oknodo $DAEMON_VERBOSITY \
				--name $WPA_SUP_PNAME --startas $WPA_SUP_BIN --pidfile $WPA_SUP_PIDFILE \
				-- $WPA_SUP_OPTIONS -D $WPA_SUP_DRIVER $WPA_SUP_CONF
		fi
		
		# loop wait until ctrl_interface socket exists for a max of 5 seconds
		# in some situations, init_wpa_cli() or conf_wpa_supplicant() were being launched
		# before wpa_supplicant had established the ctrl_interface socket
		# i cannot say for certain if this is absolutely required, but it cannot hurt to have a buffer
		local WPA_SOCKET_WAIT="0" MAX_WPA_SOCKET_WAIT="5"
		until [ -S "$WPA_CTRL_DIR/$IFACE" ]; do
			if [ "$WPA_SOCKET_WAIT" -ge "$MAX_WPA_SOCKET_WAIT" ]; then
				# socket should be created immediately after invocation
				# a few seconds is ample time even on ancient hardware
				wpa_msg stderr "ctrl_interface socket not found at $WPA_CTRL_DIR/$IFACE"
				return 1
			else
				wpa_msg verbose "waiting for \"$WPA_CTRL_DIR/$IFACE\": $WPA_SOCKET_WAIT (max. $MAX_WPA_SOCKET_WAIT)"
				WPA_SOCKET_WAIT=$(($WPA_SOCKET_WAIT+1))
				sleep 1
			fi
		done
		
		wpa_msg verbose "ctrl_interface socket located at $WPA_CTRL_DIR/$IFACE"
		return 0			
	else
		wpa_msg stderr "unknown or stale ctrl_interface socket located at $WPA_CTRL_DIR/$IFACE"
		return 1
	fi
}

conf_wpa_supplicant () {
	
	# global settings
	wpa_cli_do "$IF_WPA_AP_SCAN" raw \
		ap_scan wpa-ap-scan
	
	wpa_cli_do "$IF_WPA_PREAUTHENTICATE" raw \
		preauthenticate wpa-preauthenticate
		
	# per ssid specific settings
	if [ -n "$IF_WPA_SSID" ]; then
		
		# initialize a new network block, record the unique WPA_ID
		WPA_ID=$(wpa_cli add_network)
		
		wpa_msg verbose "configuring network block -- $WPA_ID"
		
		# set an ssid
		wpa_cli_do "$IF_WPA_SSID" ascii \
			set_network ssid wpa-ssid
		
		# configure the new network block
		wpa_cli_do "$IF_WPA_PRIORITY" raw \
			set_network priority wpa-priority
		
		wpa_cli_do "$IF_WPA_BSSID" raw \
			set_network bssid wpa-bssid
		
		if [ -s "$IF_WPA_PSK_FILE" ]; then
			IF_WPA_PSK=$(cat "$IF_WPA_PSK_FILE")
		fi
		wpa_cli_do "$IF_WPA_PSK" raw \
			set_network psk wpa-psk
		
		if [ -s "$IF_WPA_PASSPHRASE_FILE" ]; then
			IF_WPA_PASSPHRASE=$(cat "$IF_WPA_PASSPHRASE_FILE")
		fi
		wpa_cli_do "$IF_WPA_PASSPHRASE" ascii \
			set_network psk wpa-passphrase
		
		wpa_cli_do "$IF_WPA_PAIRWISE" raw \
			set_network pairwise wpa-pairwise
		
		wpa_cli_do "$IF_WPA_GROUP" raw \
			set_network group wpa-group
		
		wpa_cli_do "$IF_WPA_KEY_MGMT" raw \
			set_network key_mgmt wpa-key-mgmt
		
		wpa_cli_do "$IF_WPA_PROTO" raw \
			set_network proto wpa-proto
		
		wpa_cli_do "$IF_WPA_AUTH_ALG" raw \
			set_network auth_alg wpa-auth-alg
		
		wpa_cli_do "$IF_WPA_SCAN_SSID" raw \
			set_network scan_ssid wpa-scan-ssid
		
		wpa_cli_do "$IF_WPA_IDENTITY" ascii \
			set_network identity wpa-identity
		
		wpa_cli_do "$IF_WPA_ANONYMOUS_IDENTITY" ascii \
			set_network anonymous_identity wpa-anonymous-identity
		
		wpa_cli_do "$IF_WPA_EAP" raw \
			set_network eap wpa-eap
		
		wpa_cli_do "$IF_WPA_EAPPSK" raw \
			set_network eappsk wpa-eappsk

		wpa_cli_do "$IF_WPA_NAI" ascii \
			set_network nai wpa-nai

		wpa_cli_do "$IF_WPA_PASSWORD" ascii \
			set_network password wpa-password

		wpa_cli_do "$IF_WPA_CA_CERT" ascii \
			set_network ca_cert wpa-ca-cert

		wpa_cli_do "$IF_WPA_CA_PATH" ascii \
			set_network ca_path wpa-ca-path

		wpa_cli_do "$IF_WPA_CLIENT_CERT" ascii \
			set_network client_cert wpa-client-cert

		wpa_cli_do "$IF_WPA_PRIVATE_KEY" ascii \
			set_network private_key wpa-private-key

		wpa_cli_do "$IF_WPA_PRIVATE_KEY_PASSWD" ascii \
			set_network private_key_passwd wpa-private-key-passwd
		
		wpa_cli_do "$IF_WPA_DH_FILE" ascii \
			set_network dh_file wpa-dh-file

		wpa_cli_do "$IF_WPA_SUBJECT_MATCH" ascii \
			set_network subject_match wpa-subject-match

		wpa_cli_do "$IF_WPA_ALTSUBJECT_MATCH" ascii \
			set_network altsubject_match wpa-altsubject-match

		wpa_cli_do "$IF_WPA_CA_CERT2" ascii \
			set_network ca_cert2 wpa-ca-cert2

		wpa_cli_do "$IF_WPA_CA_PATH2" ascii \
			set_network ca_path2 wpa-ca-path2

		wpa_cli_do "$IF_WPA_CLIENT_CERT2" ascii \
			set_network client_cert2 wpa-client-cert2

		wpa_cli_do "$IF_WPA_PRIVATE_KEY2" ascii \
			set_network private_key2 wpa-private-key2

		wpa_cli_do "$IF_WPA_PRIVATE_KEY_PASSWD2" ascii \
			set_network private_key_passwd2 wpa-private-key-passwd2
		
		wpa_cli_do "$IF_WPA_DH_FILE2" ascii \
			set_network dh_file2 wpa-dh-file2

		wpa_cli_do "$IF_WPA_SUBJECT_MATCH2" ascii \
			set_network subject_match2 wpa-subject-match2

		wpa_cli_do "$IF_WPA_ALTSUBJECT_MATCH2" ascii \
			set_network altsubject_match2 wpa-altsubject-match2
		
		wpa_cli_do "$IF_WPA_EAP_METHODS" raw \
			set_network eap_methods wpa-eap-methods

		wpa_cli_do "$IF_WPA_PHASE1" ascii \
			set_network phase1 wpa-phase1

		wpa_cli_do "$IF_WPA_PHASE2" ascii \
			set_network phase2 wpa-phase2

		wpa_cli_do "$IF_WPA_PCSC" raw \
			set_network pcsc wpa-pcsc

		wpa_cli_do "$IF_WPA_PIN" ascii \
			set_network pin wpa-pin

		wpa_cli_do "$IF_WPA_ENGINE" raw \
			set_network engine wpa-engine

		wpa_cli_do "$IF_WPA_ENGINE_ID" ascii \
			set_network engine_id wpa-engine-id

		wpa_cli_do "$IF_WPA_KEY_ID" ascii \
			set_network key_id wpa-key-id

		wpa_cli_do "$IF_WPA_EAPOL_FLAGS" raw \
			set_network eapol_flags wpa-eapol-flags
		
		wpa_cli_do "$IF_WPA_WEP_KEY0" raw \
			set_network wep_key0 wpa-wep-key0
		
		wpa_cli_do "$IF_WPA_WEP_KEY1" raw \
			set_network wep_key1 wpa-wep-key1
		
		wpa_cli_do "$IF_WPA_WEP_KEY2" raw \
			set_network wep_key2 wpa-wep-key2
		
		wpa_cli_do "$IF_WPA_WEP_KEY3" raw \
			set_network wep_key3 wpa-wep-key3
		
		wpa_cli_do "$IF_WPA_WEP_TX_KEYIDX" raw \
			set_network wep_tx_keyidx wpa-wep-tx-keyidx
		
		wpa_cli_do "$IF_WPA_PROACTIVE_KEY_CACHING" raw \
			set_network proactive_key_caching wpa-proactive-key-caching
			
		wpa_cli_do "$IF_WPA_PAC_FILE" ascii \
			set_network pac_file wpa-pac-file
		
		wpa_cli_do "$IF_WPA_MODE" raw \
			set_network mode wpa-mode
		
		wpa_cli_do "$IF_WPA_STAKEY" raw \
			set_network stakey wpa-stakey
		
		wpa_cli_do "$IF_WPA_PEERKEY" raw \
			set_network peerkey wpa-peerkey
			
		wpa_cli_do "$IF_FRAGMENT_SIZE" raw \
			set_network fragment_size wpa-fragment-size

		wpa_cli_do "$IF_WPA_ID_STR" ascii \
			set_network id_str wpa-id-str
		
		# finality: activate the new network block
		wpa_cli_do "$WPA_ID" raw \
			enable_network "enabling network block"
	fi
}

kill_wpa_supplicant () {
	if test_daemon_pidfile "$WPA_SUP_BIN" "$WPA_SUP_PIDFILE"; then
		
		wpa_msg verbose "terminating $WPA_SUP_PNAME daemon via pidfile $WPA_SUP_PIDFILE"

		start-stop-daemon --stop --oknodo $DAEMON_VERBOSITY \
			--exec $WPA_SUP_BIN --pidfile $WPA_SUP_PIDFILE
	
		if [ -f "$WPA_SUP_PIDFILE" ]; then
			rm -f "$WPA_SUP_PIDFILE"
		fi
	elif [ -S "$WPA_CTRL_DIR/$IFACE" ]; then

		# terminate existing ctrl_interface socket
		wpa_msg action "terminating via ctrl_interface socket $WPA_CTRL_DIR/$IFACE"

		wpa_cli terminate >$TO_NULL
		
		# remove stale socket
		if [ -S "$WPA_CTRL_DIR/$IFACE" ]; then
			rm -f "$WPA_CTRL_DIR/$IFACE"
		fi
	fi
}

init_wpa_cli () {
	if [ -n "$WPA_ACTION_SCRIPT" ]; then

		# default wpa_cli daemon options
		# daemonise, create pidfile, supply IFACE
		WPA_CLI_OPTIONS="-B -P $WPA_CLI_PIDFILE -i $IFACE"

		# ensure the lockfile does not exist
		if [ -e "$WPA_CLI_LOCKFILE" ]; then
			rm -f "$WPA_CLI_LOCKFILE"
		fi
			
		wpa_msg verbose "$WPA_CLI_BIN $WPA_CLI_OPTIONS -p $WPA_CTRL_DIR -a $WPA_ACTION_SCRIPT"
			
		start-stop-daemon --start --oknodo $DAEMON_VERBOSITY \
			--name $WPA_CLI_PNAME --startas $WPA_CLI_BIN --pidfile $WPA_CLI_PIDFILE \
			-- $WPA_CLI_OPTIONS -p $WPA_CTRL_DIR -a $WPA_ACTION_SCRIPT
	fi
}

kill_wpa_cli () {
	if test_daemon_pidfile "$WPA_CLI_BIN" "$WPA_CLI_PIDFILE"; then
		
		wpa_msg action "sending disconnect signal to $WPA_CLI_PNAME daemon"
		
		# send DISCONNECTED signal first
		# allow wpa_action the opportunity to undo network settings
		wpa_cli disconnect >$TO_NULL

		wpa_msg verbose "terminating $WPA_CLI_PNAME daemon via pidfile $WPA_CLI_PIDFILE"
		
		start-stop-daemon --stop --oknodo $DAEMON_VERBOSITY \
			--exec $WPA_CLI_BIN --pidfile $WPA_CLI_PIDFILE
		
		if [ -f "$WPA_CLI_PIDFILE" ]; then
			rm -f "$WPA_CLI_PIDFILE"
		fi

		if [ -f "$WPA_CLI_LOCKFILE" ]; then
			rm -f "$WPA_CLI_LOCKFILE"
		fi
	fi
}

case "$MODE" in 
	start)
		case "$PHASE" in
			pre-up)
				kill_wpa_supplicant
				if init_wpa_supplicant; then
					init_wpa_cli
					conf_wpa_supplicant
				else
					exit 1
				fi
				;;
			*)
				wpa_msg stderr "unknown phase: \"$PHASE\""
				exit 1
				;;
		esac
		;;

	stop)
		case "$PHASE" in
			pre-down)
				kill_wpa_cli
				;;
			post-down)
				kill_wpa_supplicant
				;;
			*)
				wpa_msg stderr "unknown phase: \"$PHASE\""
				exit 1
				;;
		esac
		;;
	
	*)
		wpa_msg stderr "unknown mode: \"$MODE\""
		exit 1
		;;
esac
